from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List
from cpl_translation import TranslatePipe

from bot_core.configuration.server_settings import ServerSettings
from bot_core.service.message_service import MessageService
from bot_data.abc.achievement_repository_abc import AchievementRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.achievement import Achievement
from bot_data.model.user import User
from bot_data.model.user_got_achievement import UserGotAchievement
from modules.achievements.achievement_attribute_resolver import AchievementAttributeResolver
from modules.achievements.model.achievement_attribute import AchievementAttribute
from modules.base.configuration.base_server_settings import BaseServerSettings


class AchievementService:
    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        bot: DiscordBotServiceABC,
        achievements: AchievementRepositoryABC,
        users: UserRepositoryABC,
        db: DatabaseContextABC,
        message_service: MessageService,
        resolver: AchievementAttributeResolver,
        t: TranslatePipe,
    ):
        self._config = config
        self._logger = logger
        self._bot = bot
        self._achievements = achievements
        self._users = users
        self._db = db
        self._message_service = message_service
        self._t = t

        self._attributes = List(AchievementAttribute)

        self._attributes.extend(
            [
                AchievementAttribute("xp", lambda user: user.xp, "number"),
                AchievementAttribute("message_count", lambda user: user.message_count, "number"),
                AchievementAttribute("reaction_count", lambda user: user.reaction_count, "number"),
                AchievementAttribute("ontime", lambda user: user.ontime, "number"),
                AchievementAttribute("level", lambda user: user.level, "Level"),
                # special cases
                AchievementAttribute(
                    "played_on_game_server", lambda user: resolver.get_played_on_game_server(user), "GameServer"
                ),
                AchievementAttribute(
                    "last_single_ontime_hours", lambda user: resolver.get_last_ontime_hours(user), "number"
                ),
            ]
        )

        self._operators = {
            "==": lambda value, expected_value: value == expected_value,
            "!=": lambda value, expected_value: value != expected_value,
            "<=": lambda value, expected_value: value <= expected_value,
            ">=": lambda value, expected_value: value >= expected_value,
            "<": lambda value, expected_value: value < expected_value,
            ">": lambda value, expected_value: value > expected_value,
            "contains": lambda value, expected_value: expected_value in value,
        }

    def add_achievement_attribute(self, atr: AchievementAttribute):
        self._attributes.add(atr)

    def get_operators(self) -> list[str]:
        return [x for x in self._operators.keys()]

    def get_attributes(self) -> List[AchievementAttribute]:
        return self._attributes

    def _match(self, value: any, operator: str, expected_value: str) -> bool:
        return self._operators[operator](str(value), expected_value)

    def has_user_achievement_already(self, user: User, achievement: Achievement) -> bool:
        user_achievements = self._achievements.get_achievements_by_user_id(user.id)
        return user_achievements.where(lambda x: x.name == achievement.name).count() > 0

    def has_user_achievement(self, user: User, achievement: Achievement) -> bool:
        attribute: AchievementAttribute = self._attributes.where(lambda x: x.name == achievement.attribute).single()
        return self._match(attribute.resolve(user), achievement.operator, achievement.value)

    async def validate_achievements_for_user(self, user: User):
        achievements = self._achievements.get_achievements_by_server_id(user.server.id)
        for achievement in achievements:
            if self.has_user_achievement_already(user, achievement) or not self.has_user_achievement(user, achievement):
                continue

            self._achievements.add_user_got_achievement(UserGotAchievement(user, achievement, user.server))
            self._db.save_changes()
            self._give_user_xp(user)
            await self._send_achievement_notification(user.server.discord_id, user.discord_id, achievement.name)

    def _give_user_xp(self, user: User):
        settings: BaseServerSettings = self._config.get_configuration(f"BaseServerSettings_{user.server.discord_id}")
        user.xp += settings.xp_per_achievement
        self._users.update_user(user)
        self._db.save_changes()

    async def _send_achievement_notification(self, guild_id: int, member_id: int, achievement_name: str):
        member = self._bot.get_guild(guild_id).get_member(member_id)

        settings: ServerSettings = self._config.get_configuration(f"ServerSettings_{guild_id}")
        await self._message_service.send_channel_message(
            self._bot.get_channel(settings.notification_chat_id),
            self._t.transform("modules.achievements.got_new_achievement").format(member.mention, achievement_name),
            is_persistent=True,
        )
