from typing import List as TList

import discord
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.command import DiscordCommandABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List
from cpl_translation import TranslatePipe
from discord import app_commands
from discord.ext import commands
from discord.ext.commands import Context

from bot_core.abc.client_utils_abc import ClientUtilsABC
from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.helper.command_checks import CommandChecks
from bot_core.logging.command_logger import CommandLogger
from bot_data.abc.auto_role_repository_abc import AutoRoleRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.model.auto_role import AutoRole
from bot_data.model.auto_role_rule import AutoRoleRule
from modules.permission.abc.permission_service_abc import PermissionServiceABC


class AutoRoleGroup(DiscordCommandABC):
    def __init__(
        self,
        logger: CommandLogger,
        message_service: MessageServiceABC,
        bot: DiscordBotServiceABC,
        client_utils: ClientUtilsABC,
        translate: TranslatePipe,
        servers: ServerRepositoryABC,
        auto_roles: AutoRoleRepositoryABC,
        db_context: DatabaseContextABC,
        permission_service: PermissionServiceABC,
    ):
        DiscordCommandABC.__init__(self)

        self._logger = logger
        self._message_service = message_service
        self._bot = bot
        self._client_utils = client_utils
        self._t = translate
        self._servers = servers
        self._auto_roles = auto_roles
        self._db_context = db_context
        self._permissions = permission_service

        self._logger.trace(__name__, f"Loaded command service: {type(self).__name__}")

    async def _auto_role_auto_complete(
        self, interaction: discord.Interaction, current: str
    ) -> TList[app_commands.Choice[str]]:
        server = self._servers.get_server_by_discord_id(interaction.guild.id)
        auto_roles = self._auto_roles.get_auto_roles_by_server_id(server.id).select(lambda x: x.id)
        return [
            app_commands.Choice(name=auto_role, value=auto_role)
            for auto_role in self._client_utils.get_auto_complete_list(auto_roles, current)
        ]

    @commands.hybrid_group(name="auto-role")
    @commands.guild_only()
    async def auto_role(self, ctx: Context):
        pass

    @auto_role.command(alias="auto-roles")
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def list(self, ctx: Context, wait: int = None):
        self._logger.debug(__name__, f"Received command auto-role list {ctx}")

        if ctx.guild is None:
            return

        embed = discord.Embed(
            title=self._t.transform("modules.auto_role.list.title"),
            description=self._t.transform("modules.auto_role.list.description"),
            color=int("ef9d0d", 16),
        )
        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        auto_roles = self._auto_roles.get_auto_roles_by_server_id(server.id)
        if auto_roles.count() < 1:
            await self._message_service.send_ctx_msg(ctx, self._t.transform("modules.auto_role.error.nothing_found"))
            self._logger.trace(__name__, f"Finished command auto-role list")
            return

        auto_role_id = ""
        message_id = ""
        for auto_role in auto_roles:
            auto_role_id += f"\n{auto_role.id}"
            message_id += f"\n{auto_role.discord_message_id}"

        embed.add_field(
            name=self._t.transform("modules.auto_role.list.auto_role_id"),
            value=auto_role_id,
            inline=True,
        )
        embed.add_field(
            name=self._t.transform("modules.auto_role.list.message_id"),
            value=message_id,
            inline=True,
        )
        await self._message_service.send_ctx_msg(ctx, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f"Finished command auto-role list")

    @auto_role.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def add(self, ctx: Context, channel: discord.TextChannel, message_id: str):
        self._logger.debug(__name__, f"Received command auto-role add {ctx} {message_id}")

        message = (
            List(
                discord.Message,
                [message async for message in channel.history(limit=50)],
            )
            .where(lambda m: m.id == int(message_id))
            .single_or_default()
        )
        if message is None:
            self._logger.debug(__name__, f"Message with id {message_id} not found in {channel.name}")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.add.error.not_found").format(message_id, channel.mention),
            )
            self._logger.trace(__name__, f"Finished command auto-role add")
            return

        if self._auto_roles.find_auto_role_by_message_id(int(message_id)) is not None:
            self._logger.debug(__name__, f"auto-role for message {message_id} already exists")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.add.error.already_exists").format(message_id),
            )
            self._logger.trace(__name__, f"Finished command auto-role add")
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        self._auto_roles.add_auto_role(AutoRole(server, int(channel.id), int(message_id)))
        self._db_context.save_changes()
        self._logger.info(__name__, f"Saved auto-role for message {message_id} at server {server}")
        await self._message_service.send_ctx_msg(
            ctx, self._t.transform("modules.auto_role.add.success").format(message_id)
        )
        self._logger.trace(__name__, f"Finished command auto-role add")

    @add.autocomplete("message_id")
    async def add_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        channel = discord.utils.get(interaction.guild.text_channels, id=interaction.channel_id)
        try:
            channel_from_data = interaction.data["options"][0]["options"][0]["value"]
            found_channel = discord.utils.get(interaction.guild.text_channels, id=int(channel_from_data))
            if found_channel is not None:
                channel = found_channel
        finally:
            pass

        messages = [message async for message in channel.history(limit=10)]
        return [
            app_commands.Choice(name=f"{message.author}@{message.created_at}", value=str(message.id))
            for message in messages
            if current in str(message.id)
        ]

    @auto_role.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def remove(self, ctx: Context, auto_role: int):
        self._logger.debug(__name__, f"Received command auto-role remove {ctx} {auto_role}")

        auto_role_from_db = self._auto_roles.find_auto_role_by_id(auto_role)
        if auto_role_from_db is None:
            self._logger.debug(__name__, f"auto-role {auto_role} not found")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.remove.error.not_found").format(auto_role),
            )
            self._logger.trace(__name__, f"Finished command auto-role remove")
            return

        for rule in self._auto_roles.get_auto_role_rules_by_auto_role_id(auto_role_from_db.id):
            self._auto_roles.delete_auto_role_rule(rule)
            self._logger.info(__name__, f"Removed auto-role rule {rule.role_id}")

        self._auto_roles.delete_auto_role(auto_role_from_db)
        self._db_context.save_changes()
        self._logger.info(__name__, f"Removed auto-role {auto_role}")
        await self._message_service.send_ctx_msg(
            ctx, self._t.transform("modules.auto_role.remove.success").format(auto_role)
        )
        self._logger.trace(__name__, f"Finished command auto-role remove")

    @remove.autocomplete("auto_role")
    async def remove_autocomplete(
        self, interaction: discord.Interaction, current: str
    ) -> TList[app_commands.Choice[str]]:
        return await self._auto_role_auto_complete(interaction, current)

    @auto_role.group()
    @commands.guild_only()
    async def rule(self, ctx: Context):
        pass

    @rule.command(alias="rules")
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def list(self, ctx: Context, auto_role: int, wait: int = None):
        self._logger.debug(__name__, f"Received command auto-role rule list {ctx}")

        embed = discord.Embed(
            title=self._t.transform("modules.auto_role.list.title"),
            description=self._t.transform("modules.auto_role.list.description"),
            color=int("ef9d0d", 16),
        )
        rules = self._auto_roles.get_auto_role_rules_by_auto_role_id(auto_role)
        if rules.count() < 1:
            await self._message_service.send_ctx_msg(
                ctx, self._t.transform("modules.auto_role.rule.error.id_not_found")
            )
            self._logger.trace(__name__, f"Finished command auto-role rule list")
            return

        auto_role_rule_id = ""
        emoji = ""
        role = ""
        for rule in rules:
            auto_role_rule_id += f"\n{rule.role_id}"
            emoji += f"\n{rule.emoji_name}"
            role += f"\n{ctx.guild.get_role(rule.role_id)}"

        embed.add_field(
            name=self._t.transform("modules.auto_role.rule.list.auto_role_rule_id"),
            value=auto_role_rule_id,
            inline=True,
        )
        embed.add_field(
            name=self._t.transform("modules.auto_role.rule.list.emoji"),
            value=emoji,
            inline=True,
        )
        embed.add_field(
            name=self._t.transform("modules.auto_role.rule.list.role"),
            value=role,
            inline=True,
        )
        await self._message_service.send_ctx_msg(ctx, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f"Finished command auto-role rule list")

    @list.autocomplete("auto_role")
    async def list_autocomplete(
        self, interaction: discord.Interaction, current: str
    ) -> TList[app_commands.Choice[str]]:
        return await self._auto_role_auto_complete(interaction, current)

    @rule.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def add(self, ctx: Context, auto_role: int, emoji_name: str, role_id: str):
        self._logger.debug(__name__, f"Received command auto-role add {ctx} {auto_role}")

        emoji = discord.utils.get(self._bot.emojis, name=emoji_name)
        if emoji is None:
            self._logger.debug(__name__, f"auto-role rule add emoji {emoji_name} not found")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.rule.add.error.emoji_not_found").format(emoji_name, auto_role),
            )
            self._logger.trace(__name__, f"Finished command auto-role add")
            return

        role = ctx.guild.get_role(int(role_id))
        if role is None:
            self._logger.debug(__name__, f"auto-role rule add role {role_id} not found")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.rule.add.error.role_not_found").format(role_id, auto_role),
            )
            self._logger.trace(__name__, f"Finished command auto-role add")
            return

        auto_role_from_db = self._auto_roles.get_auto_role_by_id(auto_role)
        if auto_role_from_db is None:
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.remove.error.not_found").format(auto_role),
            )
            self._logger.trace(__name__, f"Finished command auto-role rule add")
            return

        if (
            self._auto_roles.get_auto_role_rules_by_auto_role_id(auto_role)
            .where(lambda r: r.emoji_name == emoji.name and int(role_id) == role.id)
            .first_or_default()
            is not None
        ):
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.add.error.already_exists").format(auto_role),
            )
            self._logger.trace(__name__, f"Finished command auto-role rule add")
            return

        self._auto_roles.add_auto_role_rule(AutoRoleRule(auto_role_from_db, emoji_name, int(role_id)))
        self._db_context.save_changes()
        rule = (
            self._auto_roles.get_auto_role_rules_by_auto_role_id(auto_role)
            .where(lambda r: r.emoji_name == emoji.name and int(role_id) == role.id)
            .single()
        )

        # as task to run in background
        self._bot.loop.create_task(
            await self._client_utils.react_to_message_by_auto_role_rule(
                auto_role_from_db.discord_channel_id, auto_role_from_db.discord_message_id, rule, ctx.guild
            )
        )

        await self._message_service.send_ctx_msg(
            ctx,
            self._t.transform("modules.auto_role.rule.add.success").format(emoji, role.mention, auto_role),
        )
        self._logger.trace(__name__, f"Finished command auto-role rule add")

    @add.autocomplete("auto_role")
    async def add_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        return await self._auto_role_auto_complete(interaction, current)

    @add.autocomplete("emoji_name")
    async def add_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        emojis = List(discord.Emoji, interaction.guild.emojis)
        return [
            app_commands.Choice(name=emoji.name, value=emoji.name)
            for emoji in self._client_utils.get_auto_complete_list(emojis, current, lambda e: e.name)
        ]

    @add.autocomplete("role_id")
    async def rule_add_autocomplete(
        self, interaction: discord.Interaction, current: str
    ) -> TList[app_commands.Choice[str]]:
        roles = List(discord.Role, interaction.guild.roles)
        return [
            app_commands.Choice(name=role.name, value=str(role.id))
            for role in self._client_utils.get_auto_complete_list(roles, current, lambda r: r.name)
        ]

    @rule.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def remove(self, ctx: Context, auto_role_rule: int):
        self._logger.debug(__name__, f"Received command auto-role remove {ctx} {auto_role_rule}")

        auto_role_from_db = self._auto_roles.get_auto_role_rule_by_id(auto_role_rule)
        if auto_role_from_db is None:
            self._logger.debug(__name__, f"auto-role rule {auto_role_rule} not found")
            await self._message_service.send_ctx_msg(
                ctx,
                self._t.transform("modules.auto_role.remove.error.not_found").format(auto_role_rule),
            )
            self._logger.trace(__name__, f"Finished command auto-role rule remove")
            return

        self._auto_roles.delete_auto_role_rule(auto_role_from_db)
        self._db_context.save_changes()
        self._logger.info(__name__, f"Removed auto-role rule {auto_role_rule}")
        await self._message_service.send_ctx_msg(
            ctx,
            self._t.transform("modules.auto_role.rule.remove.success").format(auto_role_rule),
        )
        self._logger.trace(__name__, f"Finished command auto-role remove")

    @remove.autocomplete("auto_role_rule")
    async def remove_autocomplete(
        self, interaction: discord.Interaction, current: str
    ) -> TList[app_commands.Choice[str]]:
        server = self._servers.get_server_by_discord_id(interaction.guild.id)
        auto_roles = self._auto_roles.get_auto_roles_by_server_id(server.id).select(lambda x: x.id)
        rules = auto_roles.select_many(lambda ar: self._auto_roles.get_auto_role_rules_by_auto_role_id(ar))
        return [
            app_commands.Choice(
                name=f"{rule.id} {rule.emoji_name} {interaction.guild.get_role(int(rule.role_id))}",
                value=rule.id,
            )
            for rule in self._client_utils.get_auto_complete_list(rules, current, lambda r: r.id)
        ]
