import { Component, OnDestroy, OnInit } from "@angular/core";
import { Achievement, AchievementAttribute, AchievementFilter } from "../../../../../../models/data/achievement.model";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { Subject, throwError } from "rxjs";
import { Server } from "../../../../../../models/data/server.model";
import { UserDTO } from "../../../../../../models/auth/auth-user.dto";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { TranslateService } from "@ngx-translate/core";
import { DataService } from "../../../../../../services/data/data.service";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { ActivatedRoute } from "@angular/router";
import { AchievementListQuery, AchievementTypeQuery, GameServerListQuery, LevelListQuery, Query } from "../../../../../../models/graphql/query.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { LazyLoadEvent, MenuItem } from "primeng/api";
import { Table } from "primeng/table";
import { User } from "../../../../../../models/data/user.model";
import { AchievementMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";

@Component({
  selector: "app-achievement",
  templateUrl: "./achievement.component.html",
  styleUrls: ["./achievement.component.scss"]
})
export class AchievementComponent implements OnInit, OnDestroy {
  public achievements: Achievement[] = [];
  public loading = true;

  public isEditingNew: boolean = false;

  public filterForm!: FormGroup<{
    id: FormControl<number | null>,
    name: FormControl<string | null>,
    color: FormControl<string | null>,
    min_xp: FormControl<number | null>,
    permissions: FormControl<string | null>,
  }>;

  public filter: AchievementFilter = {};
  public page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  public sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  public totalRecords: number = 0;

  public clonedAchievements: { [s: string]: Achievement; } = {};

  private unsubscriber = new Subject<void>();
  private server: Server = {};
  public user: UserDTO | null = null;

  public operators: string[] = [];
  public attributes: MenuItem[] = [];
  private achievementsAttributes: AchievementAttribute[] = [];

  public levels!: MenuItem[];
  public gameServers!: MenuItem[];

  query: string = Queries.achievementWithHistoryQuery;

  public constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute) {
  }

  public ngOnInit(): void {
    this.loading = true;
    this.setFilterForm();
    this.data.getServerFromRoute(this.route).then(async server => {
      this.server = server;
      let authUser = await this.authService.getLoggedInUser();
      this.user = authUser?.users?.find(u => u.server == this.server.id) ?? null;
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  private loadLevels() {
    this.data.query<LevelListQuery>(Queries.levelQuery, {
        serverId: this.server.id
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.levels = data.levels.map(level => {
        return { label: level.name, value: level.name };
      });
    });
  }

  private loadGameServers() {
    this.data.query<GameServerListQuery>(Queries.gameServerQuery, {
        serverId: this.server.id
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.gameServers = data.gameServers.map(gameServer => {
        return { label: gameServer.name, value: gameServer.name };
      });
    });
  }

  private loadNextData() {
    this.data.query<AchievementListQuery>(Queries.achievementQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.achievementCount;
      this.achievements = data.achievements;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  public loadNextPage(): void {
    this.data.query<AchievementTypeQuery>(Queries.achievementTypeQuery
    ).subscribe(data => {
      this.operators = data.achievementOperators;
      this.achievementsAttributes = data.achievementAttributes;
      this.attributes = data.achievementAttributes.map(attribute => {
        return { label: attribute.name, value: attribute.name };
      });
      this.loadLevels();
      this.loadGameServers();
      this.loadNextData();
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      name: new FormControl<string | null>(null),
      color: new FormControl<string | null>(null),
      min_xp: new FormControl<number | null>(null),
      permissions: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.name) {
        this.filter.name = changes.name;
      } else {
        this.filter.name = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public newAchievementTemplate: Achievement = {
    id: 0,
    createdAt: "",
    modifiedAt: ""
  };

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, user: User, index: number): void {
    this.clonedAchievements[index] = { ...user };
  }

  public onRowEditSave(table: Table, newAchievement: Achievement, index: number): void {
    if (this.isEditingNew && JSON.stringify(newAchievement) === JSON.stringify(this.newAchievementTemplate)) {
      this.isEditingNew = false;
      this.achievements.splice(index, 1);
      return;
    }

    if (!newAchievement.id && !this.isEditingNew || !newAchievement.name && !newAchievement.attribute && !newAchievement?.operator && !newAchievement?.value) {
      return;
    }

    if (this.isEditingNew) {
      this.spinner.showSpinner();
      this.data.mutation<AchievementMutationResult>(Mutations.createAchievement, {
          name: newAchievement.name,
          description: newAchievement.description,
          attribute: newAchievement.attribute,
          operator: newAchievement.operator,
          value: newAchievement.value + "",
          serverId: this.server.id
        }
      ).pipe(catchError(err => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.error(this.translate.instant("view.server.achievements.message.achievement_create_failed"), this.translate.instant("view.server.achievements.message.achievement_create_failed_d"));
        return throwError(err);
      })).subscribe(result => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.success(this.translate.instant("view.server.achievements.message.achievement_create"), this.translate.instant("view.server.achievements.message.achievement_create_d", { name: result.achievement.createAchievement?.name }));
        this.loadNextPage();
      });
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<AchievementMutationResult>(Mutations.updateAchievement, {
        id: newAchievement.id,
        name: newAchievement.name,
        description: newAchievement.description,
        attribute: newAchievement.attribute,
        operator: newAchievement.operator,
        value: newAchievement.value + ""
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      this.toastService.error(this.translate.instant("view.server.achievements.message.achievement_update_failed"), this.translate.instant("view.server.achievements.message.achievement_update_failed_d", { name: newAchievement.name }));
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.achievements.message.achievement_update"), this.translate.instant("view.server.achievements.message.achievement_update_d", { name: newAchievement.name }));
      this.loadNextPage();
    });

  }

  public onRowEditCancel(index: number): void {
    if (this.isEditingNew) {
      this.achievements.splice(index, 1);
      delete this.clonedAchievements[index];
      this.isEditingNew = false;
      return;
    }

    this.achievements[index] = this.clonedAchievements[index];
    delete this.clonedAchievements[index];
  }

  public deleteAchievement(achievement: Achievement): void {
    this.confirmDialog.confirmDialog(
      this.translate.instant("view.server.achievements.message.achievement_delete"), this.translate.instant("view.server.achievements.message.achievement_delete_q", { name: achievement.name }),
      () => {
        this.spinner.showSpinner();
        this.data.mutation<AchievementMutationResult>(Mutations.deleteAchievement, {
            id: achievement.id
          }
        ).pipe(catchError(err => {
          this.spinner.hideSpinner();
          this.toastService.error(this.translate.instant("view.server.achievements.message.achievement_delete_failed"), this.translate.instant("view.server.achievements.message.achievement_delete_failed_d", { name: achievement.name }));
          return throwError(err);
        })).subscribe(l => {
          this.spinner.hideSpinner();
          this.toastService.success(this.translate.instant("view.server.achievements.message.achievement_deleted"), this.translate.instant("view.server.achievements.message.achievement_deleted_d", { name: achievement.name }));
          this.loadNextPage();
        });
      });
  }

  public addAchievement(table: Table): void {
    const newAchievement = JSON.parse(JSON.stringify(this.newAchievementTemplate));
    newAchievement.id = Math.max.apply(Math, this.achievements.map(l => {
      return l.id ?? 0;
    })) + 1;

    this.achievements.push(newAchievement);

    table.initRowEdit(newAchievement);

    const index = this.achievements.findIndex(l => l.id == newAchievement.id);
    this.onRowEditInit(table, newAchievement, index);

    this.isEditingNew = true;
  }

  public getAchievementAttributeByName(name: string): AchievementAttribute {
    const [found] = this.achievementsAttributes.filter(x => x.name === name);
    return found;
  }
}
