import { Component, OnDestroy, OnInit, Query } from "@angular/core";
import { DataService } from "../../../../../../services/data/data.service";
import { ActivatedRoute, Router } from "@angular/router";
import { AutoRoleRule, AutoRoleRuleFilter } from "../../../../../../models/data/auto_role.model";
import { Guild } from "../../../../../../models/data/discord.model";
import { LazyLoadEvent, MenuItem } from "primeng/api";
import { User } from "../../../../../../models/data/user.model";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { TranslateService } from "@ngx-translate/core";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { AutoRoleRuleQuery, SingleDiscordQuery } from "../../../../../../models/graphql/query.model";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { Server } from "../../../../../../models/data/server.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { Table } from "primeng/table";
import { AutoRoleMutationResult, AutoRoleRuleMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";
import { Subject, throwError } from "rxjs";

@Component({
  selector: "app-auto-roles-rules",
  templateUrl: "./auto-roles-rules.component.html",
  styleUrls: ["./auto-roles-rules.component.scss"]
})
export class AutoRolesRulesComponent implements OnInit, OnDestroy {

  rules: AutoRoleRule[] = [];
  guild: Guild = { channels: [], emojis: [], roles: [] };
  emojis: MenuItem[] = [];
  roles: MenuItem[] = [];
  loading = true;

  autoRoleId!: number;

  clonedUsers: { [s: string]: User; } = {};
  isEditingNew: boolean = false;

  newAutoRoleTemplate: AutoRoleRule = {
    createdAt: "",
    modifiedAt: ""
  };

  filterForm!: FormGroup<{
    id: FormControl<number | null>,
    emojiName: FormControl<string | null>,
    roleId: FormControl<string | null>,
  }>;

  filter: AutoRoleRuleFilter = {};
  page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  totalRecords: number = 0;
  private unsubscriber = new Subject<void>();
  private server: Server = {};

  query: string = Queries.autoRoleRulesHistoryQuery;

  constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute,
    private router: Router
  ) {
  }

  public getEmojiUrl(name: string): string {
    return this.guild?.emojis.filter(x => x.name == name)[0].url ?? "";
  }

  public ngOnInit(): void {
    this.setFilterForm();
    this.data.getServerFromRoute(this.route).then(server => {
      this.server = server;
      this.spinner.showSpinner();
      if (!this.route.snapshot.params["autoRoleId"]) {
        this.spinner.hideSpinner();
        this.router.navigate(["../"]);
        return;
      }
      this.autoRoleId = +this.route.snapshot.params["autoRoleId"];

      this.spinner.showSpinner();
      this.data.query<SingleDiscordQuery>(Queries.guildsQuery, {
          filter: {
            id: server.discordId
          }
        }
      ).subscribe(data => {
        this.guild = data.guilds[0];
        this.emojis = this.guild.emojis
          .map(x => {
            return { label: x.name, value: x };
          });
        this.roles = this.guild.roles
          .map(x => {
            return { label: x.name, value: x };
          });
        this.spinner.hideSpinner();
      });
      this.loadNextPage();
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public loadNextPage(): void {
    this.loading = true;
    this.data.query<AutoRoleRuleQuery>(Queries.autoRoleRulesQuery, {
        serverId: this.server.id, autoRoleId: this.autoRoleId, filter: this.filter, page: this.page, sort: this.sort
      },
      (x: { servers: Server[] }) => {
        if (!x.servers[0].autoRoles || x.servers[0].autoRoles?.length == 0) {
          return undefined;
        }
        return x.servers[0].autoRoles[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.autoRoleRuleCount;
      this.rules = data.autoRoleRules;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      emojiName: new FormControl<string | null>(null),
      roleId: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.emojiName) {
        this.filter.emojiName = changes.emojiName;
      } else {
        this.filter.emojiName = undefined;
      }

      if (changes.roleId) {
        this.filter.roleId = changes.roleId;
      } else {
        this.filter.roleId = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, autoRoleRule: AutoRoleRule, index: number): void {
    this.clonedUsers[index] = { ...autoRoleRule };
  }

  public onRowEditSave(table: Table, newAutoRoleRule: AutoRoleRule, index: number): void {
    if (this.isEditingNew && JSON.stringify(newAutoRoleRule) === JSON.stringify(this.newAutoRoleTemplate)) {
      this.isEditingNew = false;
      this.rules.splice(index, 1);
      return;
    }

    if (!newAutoRoleRule.id && !this.isEditingNew || !newAutoRoleRule.emojiName && !newAutoRoleRule.roleId) {
      return;
    }

    if (this.isEditingNew) {
      this.spinner.showSpinner();
      this.data.mutation<AutoRoleRuleMutationResult>(Mutations.createAutoRoleRule, {
          autoRoleId: this.autoRoleId,
          emojiName: newAutoRoleRule.emojiName,
          roleId: newAutoRoleRule.roleId
        }
      ).pipe(catchError(err => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.error(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_create_failed"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_create_failed_d"));
        return throwError(err);
      })).subscribe(result => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.success(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_create"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_create_d", { id: result.autoRoleRule.createAutoRoleRule?.id }));
        this.loadNextPage();
      });
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<AutoRoleMutationResult>(Mutations.updateAutoRoleRule, {
        id: newAutoRoleRule.id,
        emojiName: newAutoRoleRule.emojiName,
        roleId: newAutoRoleRule.roleId
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      this.toastService.error(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_update_failed"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_update_failed_d"));
      return throwError(err);
    })).subscribe(result => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_updated"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_update_d", { id: result.autoRole.createAutoRole?.id }));
      this.loadNextPage();
    });
  }

  public onRowEditCancel(index: number): void {
    if (this.isEditingNew) {
      this.rules.splice(index, 1);
      delete this.clonedUsers[index];
      this.isEditingNew = false;
      return;
    }

    this.rules[index] = this.clonedUsers[index];
    delete this.clonedUsers[index];
  }

  public deleteAutoRoleRule(autoRoleRule: AutoRoleRule): void {
    this.confirmDialog.confirmDialog(
      this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_delete"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_delete_q", { id: autoRoleRule.id }),
      () => {
        this.spinner.showSpinner();
        this.data.mutation<AutoRoleMutationResult>(Mutations.deleteAutoRoleRule, {
            id: autoRoleRule.id
          }
        ).pipe(catchError(err => {
          this.spinner.hideSpinner();
          this.toastService.error(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_delete_failed"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_delete_failed_d", { id: autoRoleRule.id }));
          return throwError(err);
        })).subscribe(_ => {
          this.spinner.hideSpinner();
          this.toastService.success(this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_deleted"), this.translate.instant("view.server.auto_roles.rules.message.auto_role_rule_deleted_d", { id: autoRoleRule.id }));
          this.loadNextPage();
        });
      });
  }

  public addAutoRoleRule(table: Table): void {
    const newAutoRole = JSON.parse(JSON.stringify(this.newAutoRoleTemplate));
    newAutoRole.id = Math.max.apply(Math, this.rules.map(u => {
      return u.id ?? 0;
    })) + 1;

    this.rules.push(newAutoRole);

    table.initRowEdit(newAutoRole);

    const index = this.rules.findIndex(u => u.id == newAutoRole.id);
    this.onRowEditInit(table, newAutoRole, index);

    this.isEditingNew = true;
  }

}
